import java.applet.Applet;
//import java.awt.event.*;
import java.awt.*;
import java.awt.image.ColorModel;
import java.awt.image.MemoryImageSource;
import java.lang.InterruptedException;
import java.util.Vector;
import java.net.URL;
import java.awt.image.PixelGrabber;
import java.net.MalformedURLException;

public class PictureTest extends Applet implements Runnable
{
    // ------------------------------------------------
    // input from HTML
    // ------------------------------------------------
    Vector images = null;    // list of images
    URL hrefURL = null;      // link to on click
    URL imageSource = null;  // folder to search for images
    MediaTracker tracker;    // used for loading
    boolean error;

    static final String imageLabel = "image";
    static final String appName = "lightbox";
    
    // ------------------------------------------------
    // used for drawing to the screen
    // ------------------------------------------------
    boolean dirty;
    
    Image img;
    Image offscreen = null;
    Graphics offscreeng = null;

    // ------------------------------------------------
    // general application
    // ------------------------------------------------
    ThreadGroup appletThreadGroup;
    Thread imagemuncher;

    // ------------------------------------------------
    // from animator applet
    // ------------------------------------------------
    Vector parseImages(String attr, String pattern)
	throws MalformedURLException {
	    if (pattern == null) {
		pattern = "light%N.jpg";
	    }
	    Vector result = new Vector(10);
	    for (int i = 0; i < attr.length(); ){
		int next = attr.indexOf('|', i);
		if (next == -1) next = attr.length();
		String file = attr.substring(i, next);
		result.addElement(new URL(imageSource, doSubst(pattern, file)));
		i = next + 1;
	    }
	    return result;
    }

    void tellLoadingMsg(String file, String fileType) {
	showStatus(appName + ": loading "+fileType+" "+file);
    }

    void tellLoadingMsg(URL url, String fileType) {
	tellLoadingMsg(url.toExternalForm(), fileType);
    }

    void clearLoadingMessage() {
	showStatus("");
    }

    boolean fetchImages(Vector images) {
	if (images == null) {
	    return true;
	}

	int size = images.size();
	for (int i = 0; i < size; i++) {
	    Object o = images.elementAt(i);
	    if (o instanceof URL) {
		URL url = (URL)o;
		tellLoadingMsg(url, imageLabel);
		Image im = getImage(url);
		tracker.addImage(im, 2);
		images.setElementAt(im, i);
	    }
	}

	try {
	    tracker.waitForID(2);
	} catch (InterruptedException e) {}
	if (tracker.isErrorID(2)) {
	    return false;
	}

	return true;
	
    }

    /**
     * Substitute an integer some number of times in a string, subject to
     * parameter strings embedded in the string.
     * Parameter strings:
     *   %N - substitute the integer as is, with no padding.
     *   %<digit>, for example %5 - substitute the integer left-padded with
     *        zeros to <digits> digits wide.
     *   %% - substitute a '%' here.
     * @param inStr the String to substitute within
     * @param theInt the int to substitute, as a String.
     */
    String doSubst(String inStr, String theInt) {
	String padStr = "0000000000";
	int length = inStr.length();
	StringBuffer result = new StringBuffer(length);
	
	for (int i = 0; i < length;) {
	    char ch = inStr.charAt(i);
	    if (ch == '%') {
		i++;
		if (i == length) {
		    result.append(ch);
		} else {
		    ch = inStr.charAt(i);
		    if (ch == 'N' || ch == 'n') {
			// just stick in the number, unmolested
			result.append(theInt);
			i++;
		    } else {
			int pad;
			if ((pad = Character.digit(ch, 10)) != -1) {
			    // we've got a width value
			    String numStr = theInt;
			    String scr = padStr+numStr;
			    result.append(scr.substring(scr.length() - pad));
			    i++;
			} else {
			    result.append(ch);
			    i++;
			}
		    }
		}
	    } else {
		result.append(ch);
		i++;
	    }
	}
	return result.toString();
    }	

public String getParam(String key) {
    String result = getParameter(key);
    //dbg("getParameter("+key+") = "+result);
    return result;
}
    
    
    /**
     * Stuff a range of image names into a Vector.
     *
     * @return a Vector of image URLs.
     */
    Vector prepareImageRange(int startImage, int endImage, String pattern)
	throws MalformedURLException {
	    
	Vector result = new Vector(Math.abs(endImage - startImage) + 1);
	if (pattern == null) {
	    pattern = "light%N.jpg";
	}
	if (startImage > endImage) {
	    for (int i = startImage; i >= endImage; i--) {
		result.addElement(new URL(imageSource, doSubst(pattern, i+"")));
	    }
	} else {
	    for (int i = startImage; i <= endImage; i++) {
		result.addElement(new URL(imageSource, doSubst(pattern, i+"")));
	    }
	}

	return result;
    }
    
    
    // ------------------------------------------------
    // implementation
    // ------------------------------------------------
    public void init() {

	tracker = new MediaTracker(this);
	
	try {
	    String param = getParam("IMAGESOURCE");	
	    imageSource = (param == null) ? getDocumentBase() : new URL(getDocumentBase(), param + "/");
	    
	    String href = getParam("HREF");
	    if (href != null) {
		try {
		    hrefURL = new URL(getDocumentBase(), href);
		} catch (MalformedURLException e) {
		    showParseError(e);
		}
	    }
	    
	    int startImage = 1;
	    int endImage = 1;
	    
	    param = getParam("ENDIMAGE");
	    if (param != null) {
		endImage = Integer.parseInt(param);
		param = getParam("STARTIMAGE");
		if (param != null) {
		    startImage = Integer.parseInt(param);
		}
		param = getParam("NAMEPATTERN");
		images = prepareImageRange(startImage, endImage, param);
	    } else {
		param = getParam("STARTIMAGE");
		if (param != null) {
		    startImage = Integer.parseInt(param);
		    param = getParam("NAMEPATTERN");
		    images = prepareImageRange(startImage, endImage, param);
		} else {
		    param = getParam("IMAGES");
		    if (param == null) {
			showStatus("No legal IMAGES, STARTIMAGE, or ENDIMAGE "+
				   "specified.");
			error = true;
			return;
		    } else {
			images = parseImages(param, getParam("NAMEPATTERN"));
		    }
		}
	    }
	} catch (MalformedURLException e) {
	    showParseError(e);
	}

	appletThreadGroup = Thread.currentThread().getThreadGroup();
    }

    void showParseError(Exception e) {
	String errorMsg = appName + ": Parse error: "+e;
	showStatus(errorMsg);
	System.err.println(errorMsg);
	error = true;
	repaint();
    }
    	
    public synchronized void start() {
	if (img == null) {
	    imagemuncher = new Thread(appletThreadGroup, this);
	    imagemuncher.start();
	}
    }

    public synchronized void stop() {
	try {
	    if (imagemuncher != null) {
		imagemuncher.stop();
	    }
	} catch (Exception e) {

	}
	imagemuncher = null;
    }

    public void restart() {
	stop();
	img = null;
	start();
    }

    public void update(Graphics g) {
	if (dirty) paint(g);
    }
    
    public void paint(Graphics g) {
	int w = getSize().width;
	int h = getSize().height;

	if (offscreen == null) {
	    // setup the offscreen
	    offscreen = createImage(w, h);
	    offscreeng = offscreen.getGraphics();
	}

	// yep
	System.gc();
	
	if (img != null) {
	    offscreeng.drawImage(img, 0, 0, this);
	    g.drawImage(offscreen, 0, 0, this);
	}

	dirty = false;
	
    }

    
    public static void main(String args[]) {
	Frame f = new Frame("2D Speed test");
	PictureTest	PictureTest = new PictureTest();

	PictureTest.init();
	PictureTest.start();
    }

    public void run() {

	Thread self = Thread.currentThread();
	if (self != imagemuncher) return;
	
	self.setPriority(3);

	if (!fetchImages(images)) {
	    // get images error
	    Object errors[] = tracker.getErrorsAny();
	    String errorMsg = "Image load error";
	    showStatus(errorMsg);
	    System.err.println(errorMsg);
	    error = true;
	    repaint();
	    return;
	}
	
	int width = getSize().width;
	int height = getSize().height;

	// ____ pixel grabbers for the images
	int size = images.size();
	int[][] grabpixels = new int[size][width * height];
	
	for (int s = 0; s < size; s++) {
	    PixelGrabber temp = new PixelGrabber((Image)images.elementAt(s),
						 0, 0, width, height, grabpixels[s], 0, width);
	    try {
		temp.grabPixels();
	    } catch (InterruptedException e) {
		System.err.println("interrupted waiting for pixels!");
		return;
	    }
	}

	int pixels[] = new int[width * height];
	int scan[]   = new int[width];

	float pi2 = 3.141592653589f / 2;
	
	int z = 0;
	while (true) {
	    
	    int index = 0;
	    int dp[] = new int[size];

	    dp[0] = 255;

	    float f1 = 32.0f;
	    float f2 = 50.0f;
	    
	    if (z > 63) dp[1] = (int)(Math.sin((float)(z - 63)/f1 - pi2) * 63 + 64);
	    else dp[1] = 0;
	    
	    if (z > 127) dp[2] = (int)(Math.sin((float)(z - 127)/f2 - pi2) * 63 + 64);
	    else dp[2] = 0;

	    for (int j = 0; j < height; j++) {

		// loop images, then scanlines (to make inner loop faster)
		for (int im = 0; im < size; im ++) {
		    
		    if (im == 0) {
			// zero the scanline: make this faster
			for (int i = 0; i < width; i++) {
			    scan[i] = 0;
			}
		    }

		    int dpx = dp[im];
		    int readindex = j * width;
		    
		    for (int i = 0; i < width; i++) {
			int p = grabpixels[im][readindex++];

			int r = p & 0xff0000;
			int g = p & 0xff00;
			int b = p & 0xff;

			r = (r * dpx >> 8) & 0xff0000;
			g = (g * dpx >> 8) & 0xff00;
			b = (b * dpx >> 8) & 0xff;

			int co = r | g | b;
			int sc = scan[i];

			// saturated add
			//int sum = sc + ~co;
			//int cbt = sc ^ ~co;

			//cbt = (cbt ^ sum) & 0x01010100;
			//sc = (co & ~0x01010100) | (sc & 0x01010100);

			sc = co + sc;
			
			scan[i] = 0xff000000 | sc;
		    }
		}
		
		// ____ now copy (use copyArray??!)
		for (int i = 0; i < width; i++) {
		    pixels[index++] = scan[i];
		}
	    }

	    z ++;

	    dirty = true;
	    
	    // once per frame
	    updateImage(self, width, height, pixels);
	}
    }

    synchronized void updateImage(Thread me, int width, int height,
			       int pixels[]) {
	
	img = createImage(new MemoryImageSource(width, height,
						ColorModel.getRGBdefault(),
						pixels, 0, width));
	repaint();
    }
}


